var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    function adopt(value) { return value instanceof P ? value : new P(function (resolve) { resolve(value); }); }
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : adopt(result.value).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
import { isArray, cloneDeep } from 'lodash-es';
import { Observable } from 'rxjs';
import { getDeviceId } from '../device';
import { Aggregation } from '../aggregation';
import { Query } from '../query';
import { KinveyError } from '../errors/kinvey';
import { getSession, formatKinveyBaasUrl, HttpRequestMethod, KinveyHttpRequest, KinveyBaasNamespace, KinveyHttpAuth } from '../http';
import { getAppKey, getApiVersion } from '../kinvey';
import { subscribeToChannel, unsubscribeFromChannel } from '../live';
import { Kmd } from '../kmd';
export function createRequest(method, url, body) {
    return new KinveyHttpRequest({
        method,
        auth: KinveyHttpAuth.SessionOrMaster,
        url,
        body
    });
}
export class NetworkStore {
    constructor(collectionName) {
        this.collectionName = collectionName;
    }
    get collection() {
        return this.collectionName;
    }
    get pathname() {
        return `/${this.collectionName}`;
    }
    get channelName() {
        return `${getAppKey()}.c-${this.collectionName}`;
    }
    getPersonalChannelName() {
        return __awaiter(this, void 0, void 0, function* () {
            const session = yield getSession();
            if (session) {
                return `${this.channelName}.u-${session._id}`;
            }
            return undefined;
        });
    }
    find(query, options = {}) {
        const stream = Observable.create((observer) => __awaiter(this, void 0, void 0, function* () {
            try {
                if (query && !(query instanceof Query)) {
                    throw new KinveyError('Invalid query. It must be an instance of the Query class.');
                }
                const { rawResponse = false, timeout, properties, trace, skipBL, kinveyFileTTL, kinveyFileTLS, } = options;
                const queryObject = Object.assign({}, query ? query.toQueryObject() : {}, { kinveyfile_ttl: kinveyFileTTL, kinveyfile_tls: kinveyFileTLS });
                const url = formatKinveyBaasUrl(KinveyBaasNamespace.AppData, this.pathname, queryObject);
                const request = createRequest(HttpRequestMethod.GET, url);
                request.headers.setCustomRequestProperties(properties);
                request.timeout = timeout;
                const response = yield request.execute();
                if (rawResponse === true) {
                    observer.next(response);
                }
                else {
                    observer.next(response.data);
                }
                observer.complete();
            }
            catch (error) {
                observer.error(error);
            }
        }));
        return stream;
    }
    count(query, options = {}) {
        const stream = Observable.create((observer) => __awaiter(this, void 0, void 0, function* () {
            try {
                if (query && !(query instanceof Query)) {
                    throw new KinveyError('Invalid query. It must be an instance of the Query class.');
                }
                const { rawResponse = false, timeout, properties, trace, skipBL } = options;
                const queryObject = Object.assign({}, query ? query.toQueryObject() : {}, {});
                const url = formatKinveyBaasUrl(KinveyBaasNamespace.AppData, `${this.pathname}/_count`, queryObject);
                const request = createRequest(HttpRequestMethod.GET, url);
                request.headers.setCustomRequestProperties(properties);
                request.timeout = timeout;
                const response = yield request.execute();
                if (rawResponse === true) {
                    observer.next(response);
                }
                else {
                    observer.next(response.data.count);
                }
                observer.complete();
            }
            catch (error) {
                observer.error(error);
            }
        }));
        return stream;
    }
    group(aggregation, options = {}) {
        const stream = Observable.create((observer) => __awaiter(this, void 0, void 0, function* () {
            try {
                if (!(aggregation instanceof Aggregation)) {
                    throw new KinveyError('Invalid aggregation. It must be an instance of the Aggregation class.');
                }
                const { rawResponse = false, timeout, properties, trace, skipBL } = options;
                const queryObject = {};
                const url = formatKinveyBaasUrl(KinveyBaasNamespace.AppData, `${this.pathname}/_group`, queryObject);
                const request = createRequest(HttpRequestMethod.POST, url, aggregation.toPlainObject());
                request.headers.setCustomRequestProperties(properties);
                request.timeout = timeout;
                const response = yield request.execute();
                if (rawResponse === true) {
                    observer.next(response);
                }
                else {
                    observer.next(response.data);
                }
                observer.complete();
            }
            catch (error) {
                observer.error(error);
            }
        }));
        return stream;
    }
    findById(id, options = {}) {
        const stream = Observable.create((observer) => __awaiter(this, void 0, void 0, function* () {
            try {
                // if (!id) {
                //   throw new Error('No id was provided. A valid id is required.');
                // }
                if (id) {
                    const { rawResponse = false, timeout, properties, trace, skipBL, kinveyFileTTL, kinveyFileTLS, } = options;
                    const queryObject = { kinveyfile_ttl: kinveyFileTTL, kinveyfile_tls: kinveyFileTLS };
                    const url = formatKinveyBaasUrl(KinveyBaasNamespace.AppData, `${this.pathname}/${id}`, queryObject);
                    const request = createRequest(HttpRequestMethod.GET, url);
                    request.headers.setCustomRequestProperties(properties);
                    request.timeout = timeout;
                    const response = yield request.execute();
                    if (rawResponse === true) {
                        observer.next(response);
                    }
                    else {
                        observer.next(response.data);
                    }
                }
                else {
                    observer.next(undefined);
                }
                observer.complete();
            }
            catch (error) {
                observer.error(error);
            }
        }));
        return stream;
    }
    create(docs, options = {}) {
        return __awaiter(this, void 0, void 0, function* () {
            const batchSize = 100;
            const apiVersion = getApiVersion();
            if (isArray(docs)) {
                if (apiVersion < 5) {
                    throw new KinveyError('Unable to create an array of entities. Please create entities one by one or use API version 5 or newer.');
                }
                if (docs.length === 0) {
                    throw new KinveyError('Unable to create an array of entities. The array must not be empty.');
                }
            }
            if (isArray(docs) && docs.length > batchSize) {
                let i = 0;
                const batchCreate = (previousResult = { entities: [], errors: [] }) => __awaiter(this, void 0, void 0, function* () {
                    if (i >= docs.length) {
                        return previousResult;
                    }
                    const batch = docs.slice(i, i + batchSize);
                    const result = yield this.create(batch, options);
                    result.errors.forEach((e) => (e.index += i));
                    const combinedResult = {
                        entities: previousResult.entities.concat(result.entities),
                        errors: previousResult.errors.concat(result.errors)
                    };
                    i += batchSize;
                    return batchCreate(combinedResult);
                });
                return batchCreate();
            }
            const { rawResponse = false, timeout, properties, trace, skipBL } = options;
            const queryObject = {};
            const url = formatKinveyBaasUrl(KinveyBaasNamespace.AppData, this.pathname, queryObject);
            const request = createRequest(HttpRequestMethod.POST, url, docs);
            request.headers.setCustomRequestProperties(properties);
            request.timeout = timeout;
            const response = yield request.execute();
            if (rawResponse === true) {
                return response;
            }
            return response.data;
        });
    }
    update(doc, options = {}) {
        return __awaiter(this, void 0, void 0, function* () {
            if (isArray(doc)) {
                throw new KinveyError('Unable to update an array of entities. Please update entities one by one.');
            }
            if (!doc._id) {
                throw new KinveyError('The entity provided does not contain an _id. An _id is required to update the entity.');
            }
            const { rawResponse = false, timeout, properties, trace, skipBL } = options;
            const queryObject = {};
            const url = formatKinveyBaasUrl(KinveyBaasNamespace.AppData, `${this.pathname}/${doc._id}`, queryObject);
            const request = createRequest(HttpRequestMethod.PUT, url, doc);
            request.headers.setCustomRequestProperties(properties);
            request.timeout = timeout;
            const response = yield request.execute();
            if (rawResponse === true) {
                return response;
            }
            return response.data;
        });
    }
    save(doc, options) {
        return __awaiter(this, void 0, void 0, function* () {
            const apiVersion = getApiVersion();
            if (apiVersion >= 5 && isArray(doc)) {
                throw new KinveyError('Unable to save an array of entities. Use "create" method to insert multiple entities.');
            }
            if (doc._id) {
                const kmd = new Kmd(cloneDeep(doc));
                if (!kmd.isLocal()) {
                    return this.update(doc, options);
                }
            }
            return this.create(doc, options);
        });
    }
    remove(query, options = {}) {
        return __awaiter(this, void 0, void 0, function* () {
            if (!query) {
                throw new KinveyError('A query must be provided to remove entities.');
            }
            if (!(query instanceof Query)) {
                throw new KinveyError('Invalid query. It must be an instance of the Query class.');
            }
            const { rawResponse = false, timeout, properties, trace, skipBL } = options;
            const queryObject = Object.assign({}, query ? query.toQueryObject() : {});
            const url = formatKinveyBaasUrl(KinveyBaasNamespace.AppData, this.pathname, queryObject);
            const request = createRequest(HttpRequestMethod.DELETE, url);
            request.headers.setCustomRequestProperties(properties);
            request.timeout = timeout;
            const response = yield request.execute();
            if (rawResponse === true) {
                return response;
            }
            return response.data;
        });
    }
    removeById(id, options = {}) {
        return __awaiter(this, void 0, void 0, function* () {
            const { rawResponse = false, timeout, properties, trace, skipBL } = options;
            const queryObject = {};
            const url = formatKinveyBaasUrl(KinveyBaasNamespace.AppData, `${this.pathname}/${id}`, queryObject);
            const request = createRequest(HttpRequestMethod.DELETE, url);
            request.headers.setCustomRequestProperties(properties);
            request.timeout = timeout;
            const response = yield request.execute();
            if (rawResponse === true) {
                return response;
            }
            return response.data;
        });
    }
    subscribe(receiver, options = {}) {
        return __awaiter(this, void 0, void 0, function* () {
            const { timeout, properties, trace, skipBL } = options;
            const deviceId = yield getDeviceId();
            const url = formatKinveyBaasUrl(KinveyBaasNamespace.AppData, `${this.pathname}/_subscribe`);
            const request = createRequest(HttpRequestMethod.POST, url, { deviceId });
            request.headers.setCustomRequestProperties(properties);
            request.timeout = timeout;
            yield request.execute();
            subscribeToChannel(this.channelName, receiver);
            const personalChannelName = yield this.getPersonalChannelName();
            if (personalChannelName) {
                subscribeToChannel(personalChannelName, receiver);
            }
            return true;
        });
    }
    unsubscribe(options = {}) {
        return __awaiter(this, void 0, void 0, function* () {
            const { timeout, properties, trace, skipBL } = options;
            const deviceId = yield getDeviceId();
            const url = formatKinveyBaasUrl(KinveyBaasNamespace.AppData, `${this.pathname}/_unsubscribe`);
            const request = createRequest(HttpRequestMethod.POST, url, { deviceId });
            request.headers.setCustomRequestProperties(properties);
            request.timeout = timeout;
            yield request.execute();
            unsubscribeFromChannel(this.channelName);
            unsubscribeFromChannel(yield this.getPersonalChannelName());
            return true;
        });
    }
}
//# sourceMappingURL=networkstore.js.map