var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    function adopt(value) { return value instanceof P ? value : new P(function (resolve) { resolve(value); }); }
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : adopt(result.value).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
import { Query } from '../query';
import { SyncError } from '../errors/sync';
import { NotFoundError } from '../errors/notFound';
import { NetworkStore } from './networkstore';
import { DataStoreCache, SyncCache, SyncEvent } from './cache';
import { getApiVersion } from '../kinvey';
const pushInProgress = new Map();
function markPushStart(collectionName) {
    pushInProgress.set(collectionName, true);
}
function markPushEnd(collectionName) {
    pushInProgress.set(collectionName, false);
    ;
}
export function queryToSyncQuery(query) {
    if (query && query instanceof Query) {
        const newFilter = Object.keys(query.filter)
            .reduce((filter, field) => Object.assign({}, filter, { [`entity.${field}`]: query.filter[field] }), {});
        const newSort = Object.keys(query.sort)
            .reduce((sort, field) => Object.assign({}, sort, { [`entity.${field}`]: query.sort[field] }), {});
        return new Query({
            filter: newFilter,
            sort: newSort,
            skip: query.skip,
            limit: query.limit
        });
    }
    return undefined;
}
export class Sync {
    constructor(collectionName, tag) {
        this.collectionName = collectionName;
        this.tag = tag;
    }
    isPushInProgress() {
        return pushInProgress.get(this.collectionName) === true;
    }
    find(providedQuery) {
        const syncCache = new SyncCache(this.tag);
        const query = new Query(providedQuery).equalTo('collection', this.collectionName);
        return syncCache.find(query);
    }
    findById(id) {
        const syncCache = new SyncCache(this.tag);
        return syncCache.findById(id);
    }
    count(providedQuery) {
        const syncCache = new SyncCache(this.tag);
        const query = new Query(providedQuery).equalTo('collection', this.collectionName);
        return syncCache.count(query);
    }
    addCreateSyncEvent(docs) {
        return this.addSyncEvent(SyncEvent.Create, docs);
    }
    addUpdateSyncEvent(docs) {
        return this.addSyncEvent(SyncEvent.Update, docs);
    }
    addDeleteSyncEvent(docs) {
        return this.addSyncEvent(SyncEvent.Delete, docs);
    }
    addSyncEvent(event, docs) {
        return __awaiter(this, void 0, void 0, function* () {
            const syncCache = new SyncCache(this.tag);
            let singular = false;
            let syncDocs = [];
            let docsToSync = docs;
            if (!Array.isArray(docs)) {
                singular = true;
                docsToSync = [docs];
            }
            if (docsToSync.length > 0) {
                const docWithNoId = docsToSync.find((doc) => !doc._id);
                if (docWithNoId) {
                    throw new SyncError('A doc is missing an _id. All docs must have an _id in order to be added to the sync collection.');
                }
                // Remove existing sync events that match the docs
                const query = new Query().contains('entityId', docsToSync.map((doc) => doc._id));
                yield this.remove(query);
                // Don't add delete events for docs that were created offline
                if (event === SyncEvent.Delete) {
                    docsToSync = docsToSync.filter((doc) => {
                        if (doc._kmd && doc._kmd.local === true) {
                            return false;
                        }
                        return true;
                    });
                }
                // Add sync events for the docs
                syncDocs = yield syncCache.save(docsToSync.map((doc) => {
                    return {
                        entityId: doc._id,
                        entity: doc,
                        collection: this.collectionName,
                        state: {
                            operation: event
                        }
                    };
                }));
            }
            return singular ? syncDocs.shift() : syncDocs;
        });
    }
    push(providedQuery, options) {
        return __awaiter(this, void 0, void 0, function* () {
            if (this.isPushInProgress()) {
                throw new SyncError('Data is already being pushed to the backend. Please wait for it to complete before pushing new data to the backend.');
            }
            const apiVersion = getApiVersion();
            const network = new NetworkStore(this.collectionName);
            const cache = new DataStoreCache(this.collectionName, this.tag);
            const syncCache = new SyncCache(this.tag);
            const collectionQuery = new Query(providedQuery).equalTo('collection', this.collectionName);
            const queryForInsert = new Query(collectionQuery).equalTo('state.operation', SyncEvent.Create);
            const syncDocsForInsert = yield syncCache.find(queryForInsert);
            const batchCreateEnabled = apiVersion >= 5 && syncDocsForInsert.length > 1;
            const totalPushResults = [];
            const batchCreateEntities = () => __awaiter(this, void 0, void 0, function* () {
                const localIdsToRemove = [];
                const entitiesForInsert = yield Promise.all(syncDocsForInsert.map((doc, index) => __awaiter(this, void 0, void 0, function* () {
                    const entity = yield cache.findById(doc.entityId);
                    if (entity._kmd && entity._kmd.local === true) {
                        localIdsToRemove[index] = doc.entityId;
                        delete entity._id;
                        delete entity._kmd.local;
                    }
                    return entity;
                })));
                let multiInsertResult;
                try {
                    multiInsertResult = yield network.create(entitiesForInsert, options);
                }
                catch (error) {
                    // In case of a general batch insert error, do not break the push operation and aggregate all errors in the result
                    if (options.catchGeneralErrors === true) {
                        return syncDocsForInsert.forEach((doc, index) => {
                            totalPushResults.push({
                                _id: doc.entityId,
                                operation: SyncEvent.Create,
                                entity: entitiesForInsert[index],
                                error
                            });
                        });
                    }
                    throw error;
                }
                // Process successful inserts
                if (multiInsertResult.entities != null) {
                    const insertedEntities = [];
                    const idsToRemoveFromSyncCache = [];
                    const idsToRemoveFromCache = [];
                    multiInsertResult.entities.forEach((insertedEntity, index) => {
                        if (insertedEntity == null) {
                            return;
                        }
                        insertedEntities.push(insertedEntity);
                        idsToRemoveFromSyncCache.push(syncDocsForInsert[index]._id);
                        idsToRemoveFromCache.push(localIdsToRemove[index]);
                        // Add the inserted entity to the end result
                        totalPushResults.push({
                            _id: syncDocsForInsert[index].entityId,
                            operation: SyncEvent.Create,
                            entity: insertedEntity
                        });
                    });
                    yield Promise.all([
                        syncCache.removeManyById(idsToRemoveFromSyncCache),
                        cache.save(insertedEntities), // Save the docs to cache
                    ]);
                    yield cache.removeManyById(idsToRemoveFromCache); // Remove the original docs that were created
                }
                // Process insert errors
                if (multiInsertResult.errors != null) {
                    multiInsertResult.errors.forEach((insertError) => {
                        // Add the error to the end result and keep the order relative to other inserts
                        totalPushResults.splice(insertError.index, 0, {
                            _id: syncDocsForInsert[insertError.index].entityId,
                            operation: SyncEvent.Create,
                            entity: entitiesForInsert[insertError.index],
                            error: insertError
                        });
                    });
                }
            });
            const createEntity = (syncDocId, entityId) => __awaiter(this, void 0, void 0, function* () {
                let doc = yield cache.findById(entityId);
                let local = false;
                try {
                    // Save the doc to the backend
                    if (doc._kmd && doc._kmd.local === true) {
                        local = true;
                        // tslint:disable-next-line:no-delete
                        delete doc._id;
                        // tslint:disable-next-line:no-delete
                        delete doc._kmd.local;
                    }
                    doc = yield network.create(doc, options);
                    // Remove the sync doc
                    yield syncCache.removeById(syncDocId);
                    // Save the doc to cache
                    yield cache.save(doc);
                    // Remove the original doc that was created
                    if (local) {
                        yield cache.removeById(entityId);
                    }
                    // Return a result
                    return {
                        _id: entityId,
                        operation: SyncEvent.Create,
                        entity: doc
                    };
                }
                catch (error) {
                    // Return a result with the error
                    return {
                        _id: entityId,
                        operation: SyncEvent.Create,
                        entity: doc,
                        error
                    };
                }
            });
            const updateEntity = (syncDocId, entityId) => __awaiter(this, void 0, void 0, function* () {
                let doc = yield cache.findById(entityId);
                try {
                    // Save the doc to the backend
                    doc = yield network.update(doc, options);
                    // Remove the sync doc
                    yield syncCache.removeById(syncDocId);
                    // Save the doc to cache
                    yield cache.save(doc);
                    // Return a result
                    return {
                        _id: entityId,
                        operation: SyncEvent.Update,
                        entity: doc
                    };
                }
                catch (error) {
                    // Return a result with the error
                    return {
                        _id: entityId,
                        operation: SyncEvent.Update,
                        entity: doc,
                        error
                    };
                }
            });
            const deleteEntity = (syncDocId, entityId) => __awaiter(this, void 0, void 0, function* () {
                try {
                    try {
                        // Remove the doc from the backend
                        yield network.removeById(entityId, options);
                    }
                    catch (error) {
                        // Rethrow the error if it is not a NotFoundError
                        if (!(error instanceof NotFoundError)) {
                            throw error;
                        }
                    }
                    // Remove the sync doc
                    yield syncCache.removeById(syncDocId);
                    // Return a result
                    return {
                        _id: entityId,
                        operation: SyncEvent.Delete
                    };
                }
                catch (error) {
                    // Return a result with the error
                    return {
                        _id: entityId,
                        operation: SyncEvent.Delete,
                        error
                    };
                }
            });
            const pushEntity = (syncDoc) => __awaiter(this, void 0, void 0, function* () {
                const { _id, entityId, state = { operation: undefined } } = syncDoc;
                switch (state.operation) {
                    case SyncEvent.Create: {
                        if (batchCreateEnabled) {
                            return null; // Inserts must have already been batched
                        }
                        return createEntity(_id, entityId);
                    }
                    case SyncEvent.Update: {
                        return updateEntity(_id, entityId);
                    }
                    case SyncEvent.Delete: {
                        return deleteEntity(_id, entityId);
                    }
                    default: {
                        return {
                            _id,
                            operation: state.operation,
                            error: new Error('Unable to push item in sync table because the event was not recognized.')
                        };
                    }
                }
            });
            // First try inserting new entities at once
            if (batchCreateEnabled) {
                try {
                    markPushStart(this.collectionName);
                    yield batchCreateEntities();
                }
                finally {
                    markPushEnd(this.collectionName);
                }
            }
            // Push other entities one by one in batches of 100 parallel requests
            const syncDocs = yield syncCache.find(collectionQuery);
            if (syncDocs.length > 0) {
                const batchSize = 100;
                let i = 0;
                const batchPush = () => __awaiter(this, void 0, void 0, function* () {
                    if (i >= syncDocs.length) {
                        return;
                    }
                    const batch = syncDocs.slice(i, i + batchSize);
                    i += batchSize;
                    try {
                        markPushStart(this.collectionName);
                        yield Promise.all(batch.map((syncDoc) => pushEntity(syncDoc)
                            .then((pushResult) => {
                            if (pushResult != null) {
                                totalPushResults.push(pushResult);
                            }
                        })));
                    }
                    finally {
                        markPushEnd(this.collectionName);
                    }
                    // Push remaining docs
                    return batchPush();
                });
                yield batchPush();
            }
            return totalPushResults;
        });
    }
    remove(providedQuery) {
        return __awaiter(this, void 0, void 0, function* () {
            const syncCache = new SyncCache(this.tag);
            const query = new Query(providedQuery).equalTo('collection', this.collectionName);
            return syncCache.remove(query);
        });
    }
    removeById(id) {
        return __awaiter(this, void 0, void 0, function* () {
            const syncCache = new SyncCache(this.tag);
            return syncCache.removeById(id);
        });
    }
    clear() {
        return __awaiter(this, void 0, void 0, function* () {
            const syncCache = new SyncCache(this.tag);
            const query = new Query().equalTo('collection', this.collectionName);
            return syncCache.remove(query);
        });
    }
}
//# sourceMappingURL=sync.js.map