var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    function adopt(value) { return value instanceof P ? value : new P(function (resolve) { resolve(value); }); }
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : adopt(result.value).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
import { isFunction, isEmpty, isArray } from 'lodash-es';
import { Base64 } from 'js-base64';
import { KinveyError } from '../errors/kinvey';
import { getAppKey, getAppSecret, getMasterSecret, getApiVersion } from '../kinvey';
import { getSession, getMFASessionToken } from './session';
function byteCount(str) {
    if (str) {
        let count = 0;
        const stringLength = str.length;
        for (let i = 0; i < stringLength; i += 1) {
            const partCount = encodeURI(str[i]).split('%').length;
            count += partCount === 1 ? 1 : partCount - 1;
        }
        return count;
    }
    return 0;
}
export class HttpHeaders {
    constructor(headers) {
        this.headers = new Map();
        this.normalizedNames = new Map();
        if (headers) {
            if (headers instanceof HttpHeaders) {
                this.join(headers);
            }
            else {
                Object.keys(headers).forEach((name) => {
                    this.set(name, headers[name]);
                });
            }
        }
    }
    get contentType() {
        return this.get('Content-Type');
    }
    set contentType(value) {
        this.set('Content-Type', value);
    }
    has(name) {
        return this.headers.has(name.toLowerCase());
    }
    get(name) {
        return this.headers.get(name.toLowerCase()) || undefined;
    }
    keys() {
        return Array.from(this.normalizedNames.values());
    }
    set(name, value) {
        if (isFunction(value)) {
            return this.set(name, value());
        }
        else if (isArray(value)) {
            return this.set(name, value.join(','));
        }
        const key = name.toLowerCase();
        this.headers.set(key, value);
        if (!this.normalizedNames.has(key)) {
            this.normalizedNames.set(key, name);
        }
        return this;
    }
    join(headers) {
        headers.keys().forEach((name) => {
            const value = headers.get(name);
            if (value) {
                this.set(name, value);
            }
        });
    }
    delete(name) {
        return this.headers.delete(name);
    }
    toPlainObject() {
        return this.keys()
            .reduce((headers, header) => {
            const value = this.get(header);
            if (value) {
                headers[header] = value;
            }
            return headers;
        }, {});
    }
}
export var KinveyHttpAuth;
(function (KinveyHttpAuth) {
    KinveyHttpAuth["All"] = "All";
    KinveyHttpAuth["App"] = "App";
    KinveyHttpAuth["Master"] = "Master";
    KinveyHttpAuth["Session"] = "Session";
    KinveyHttpAuth["SessionOrApp"] = "SessionOrApp";
    KinveyHttpAuth["SessionOrMaster"] = "SessionOrMaster";
    KinveyHttpAuth["MFASessionToken"] = "MFASessionToken";
    KinveyHttpAuth["SessionOrMFASessionTokenOrMaster"] = "SessionOrMFASessionTokenOrMaster";
})(KinveyHttpAuth || (KinveyHttpAuth = {}));
const globalHeaders = new HttpHeaders();
export function getAppVersion() {
    return globalHeaders.get('X-Kinvey-Client-App-Version');
}
export function setAppVersion(appVersion) {
    globalHeaders.set('X-Kinvey-Client-App-Version', appVersion);
}
export class KinveyHttpHeaders extends HttpHeaders {
    constructor(headers) {
        super(headers);
        // Add the Accept header
        if (!this.has('Accept')) {
            this.set('Accept', 'application/json; charset=utf-8');
        }
        // Add Content-Type header
        if (!this.has('Content-Type')) {
            this.set('Content-Type', 'application/json; charset=utf-8');
        }
        // Add the X-Kinvey-API-Version header
        if (!this.has('X-Kinvey-Api-Version')) {
            this.set('X-Kinvey-Api-Version', String(getApiVersion()));
        }
        // Add global Kinvey headers
        this.join(globalHeaders);
    }
    get requestStart() {
        return this.get('X-Kinvey-Request-Start');
    }
    setAuthorization(auth) {
        return __awaiter(this, void 0, void 0, function* () {
            const appKey = getAppKey();
            const appSecret = getAppSecret();
            const masterSecret = getMasterSecret();
            const session = yield getSession();
            const mfaSessionToken = yield getMFASessionToken();
            let value = '';
            if (auth === KinveyHttpAuth.App) {
                if (!appKey || !appSecret) {
                    throw new KinveyError('Missing appKey and/or appSecret to authorize the request.');
                }
                const credentials = Base64.encode(`${appKey}:${appSecret}`);
                value = `Basic ${credentials}`;
            }
            else if (auth === KinveyHttpAuth.Master) {
                if (!appKey || !masterSecret) {
                    throw new KinveyError('Missing appKey and/or masterSecret to authorize the request.');
                }
                const credentials = Base64.encode(`${appKey}:${masterSecret}`);
                value = `Basic ${credentials}`;
            }
            else if (auth === KinveyHttpAuth.Session) {
                if (!session || !session._kmd || !session._kmd.authtoken) {
                    throw new KinveyError('There is no active user to authorize the request.');
                }
                value = `Kinvey ${session._kmd.authtoken}`;
            }
            else if (auth === KinveyHttpAuth.MFASessionToken) {
                if (!mfaSessionToken) {
                    throw new KinveyError('Missing MFA session token to authorize the request.');
                }
                value = `KinveyMFA ${mfaSessionToken}`;
            }
            else if (auth === KinveyHttpAuth.All) {
                try {
                    return yield this.setAuthorization(KinveyHttpAuth.Session);
                }
                catch (error) {
                    try {
                        return yield this.setAuthorization(KinveyHttpAuth.App);
                    }
                    catch (error) {
                        return yield this.setAuthorization(KinveyHttpAuth.Master);
                    }
                }
            }
            else if (auth === KinveyHttpAuth.SessionOrApp) {
                try {
                    return yield this.setAuthorization(KinveyHttpAuth.Session);
                }
                catch (error) {
                    return yield this.setAuthorization(KinveyHttpAuth.App);
                }
            }
            else if (auth === KinveyHttpAuth.SessionOrMaster) {
                try {
                    return yield this.setAuthorization(KinveyHttpAuth.Session);
                }
                catch (error) {
                    return yield this.setAuthorization(KinveyHttpAuth.Master);
                }
            }
            else if (auth === KinveyHttpAuth.SessionOrMFASessionTokenOrMaster) {
                try {
                    return yield this.setAuthorization(KinveyHttpAuth.Session);
                }
                catch (error) {
                    try {
                        return yield this.setAuthorization(KinveyHttpAuth.MFASessionToken);
                    }
                    catch (error) {
                        return yield this.setAuthorization(KinveyHttpAuth.Master);
                    }
                }
            }
            this.set('Authorization', value);
        });
    }
    setCustomRequestProperties(properties) {
        const customRequestPropertiesVal = JSON.stringify(properties);
        if (!isEmpty(customRequestPropertiesVal)) {
            const customRequestPropertiesByteCount = byteCount(customRequestPropertiesVal);
            if (customRequestPropertiesByteCount >= 2000) {
                throw new Error(`The custom properties are ${customRequestPropertiesByteCount} bytes.` +
                    'It must be less then 2000 bytes.');
            }
            this.set('X-Kinvey-Custom-Request-Properties', customRequestPropertiesVal);
        }
        else {
            this.delete('X-Kinvey-Custom-Request-Properties');
        }
    }
}
//# sourceMappingURL=headers.js.map