var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    function adopt(value) { return value instanceof P ? value : new P(function (resolve) { resolve(value); }); }
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : adopt(result.value).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
import { isString } from 'lodash-es';
import PQueue from 'p-queue';
import { Base64 } from 'js-base64';
import { InvalidCredentialsError } from '../errors/invalidCredentials';
import { getAppSecret, getAppKey } from '../kinvey';
import { logger } from '../log';
import { HttpHeaders, KinveyHttpHeaders } from './headers';
import { send } from './http';
import { setSession, removeSession, getKinveyMICSession, setKinveyMICSession } from './session';
import { DataStoreCache, QueryCache, SyncCache } from '../datastore/cache';
import { formatKinveyAuthUrl, formatKinveyBaasUrl, KinveyBaasNamespace } from './utils';
const REQUEST_QUEUE = new PQueue();
let refreshTokenRequestInProgress = false;
export var HttpRequestMethod;
(function (HttpRequestMethod) {
    HttpRequestMethod["GET"] = "GET";
    HttpRequestMethod["POST"] = "POST";
    HttpRequestMethod["PUT"] = "PUT";
    HttpRequestMethod["DELETE"] = "DELETE";
})(HttpRequestMethod || (HttpRequestMethod = {}));
;
export function serialize(contentType, body) {
    if (body && !isString(body)) {
        if (contentType.indexOf('application/x-www-form-urlencoded') === 0) {
            const str = [];
            Object.keys(body).forEach((key) => {
                str.push(`${encodeURIComponent(key)}=${encodeURIComponent(body[key])}`);
            });
            return str.join('&');
        }
        else if (contentType.indexOf('application/json') === 0) {
            return JSON.stringify(body);
        }
    }
    return body;
}
export class HttpRequest {
    constructor(config) {
        this.method = HttpRequestMethod.GET;
        this.headers = new HttpHeaders(config.headers);
        if (config.method) {
            this.method = config.method;
        }
        this.url = config.url;
        this.body = config.body;
        this.timeout = config.timeout;
    }
    execute() {
        return __awaiter(this, void 0, void 0, function* () {
            // Make http request
            const response = yield send(this.toPlainObject());
            // Return the response if it was successful
            if (response.isSuccess()) {
                return response;
            }
            // Else throw the error
            throw response.error;
        });
    }
    toPlainObject() {
        return {
            headers: this.headers.toPlainObject(),
            method: this.method,
            url: this.url,
            body: this.body ? serialize(this.headers.contentType, this.body) : undefined,
            timeout: this.timeout
        };
    }
}
function isRefreshTokenRequestInProgress() {
    return refreshTokenRequestInProgress === true;
}
function markRefreshTokenRequestInProgress() {
    REQUEST_QUEUE.pause();
    refreshTokenRequestInProgress = true;
}
function markRefreshTokenRequestComplete() {
    refreshTokenRequestInProgress = false;
    REQUEST_QUEUE.start();
}
function cleanUp() {
    return __awaiter(this, void 0, void 0, function* () {
        try {
            // TODO: Unregister from live service
            // Remove the session
            yield removeSession();
            // Clear cache's
            yield QueryCache.clear();
            yield SyncCache.clear();
            yield DataStoreCache.clear();
        }
        catch (error) {
            logger.error(error.message);
        }
    });
}
function sendSessionRefreshRequest(req) {
    return __awaiter(this, void 0, void 0, function* () {
        const response = yield send(req.toPlainObject());
        if (response.statusCode > 399) {
            if (response.statusCode < 500) {
                yield cleanUp();
            }
            return null;
        }
        return response.data;
    });
}
export class KinveyHttpRequest extends HttpRequest {
    constructor(config) {
        super(config);
        this.headers = new KinveyHttpHeaders(config.headers);
        if (config.auth) {
            this.auth = config.auth;
        }
    }
    _refreshMICSession(micSession) {
        return __awaiter(this, void 0, void 0, function* () {
            const kinveyMICIdentityKey = 'kinveyAuth';
            try {
                // Refresh the MIC session
                const refreshRequest = new KinveyHttpRequest({
                    method: HttpRequestMethod.POST,
                    headers: new KinveyHttpHeaders({
                        'Content-Type': () => 'application/x-www-form-urlencoded',
                        Authorization: () => {
                            const credentials = Base64.encode(`${micSession.client_id}:${getAppSecret()}`);
                            return `Basic ${credentials}`;
                        }
                    }),
                    url: formatKinveyAuthUrl('/oauth/token'),
                    body: {
                        grant_type: 'refresh_token',
                        client_id: micSession.client_id,
                        redirect_uri: micSession.redirect_uri,
                        refresh_token: micSession.refresh_token
                    }
                });
                const refreshData = yield sendSessionRefreshRequest(refreshRequest);
                if (!refreshData) {
                    return false;
                }
                // Persist the new access and refresh tokens
                yield setKinveyMICSession(refreshData);
                const newMICSession = yield getKinveyMICSession();
                // Login with the new MIC session
                const loginRequest = new KinveyHttpRequest({
                    method: HttpRequestMethod.POST,
                    headers: new KinveyHttpHeaders({
                        'Content-Type': () => 'application/json',
                        Authorization: () => {
                            const credentials = Base64.encode(`${getAppKey()}:${getAppSecret()}`);
                            return `Basic ${credentials}`;
                        }
                    }),
                    url: formatKinveyBaasUrl(KinveyBaasNamespace.User, '/login'),
                    body: {
                        _socialIdentity: {
                            [kinveyMICIdentityKey]: newMICSession
                        }
                    }
                });
                const newSession = yield sendSessionRefreshRequest(loginRequest);
                if (!newSession) {
                    return false;
                }
                newSession._socialIdentity[kinveyMICIdentityKey] = Object.assign({}, newSession._socialIdentity[kinveyMICIdentityKey], newMICSession);
                // Set the new session
                yield setSession(newSession);
                return true;
            }
            catch (error) {
                logger.error(error.message);
                return false;
            }
        });
    }
    execute(retry = true) {
        const _super = Object.create(null, {
            execute: { get: () => super.execute }
        });
        return __awaiter(this, void 0, void 0, function* () {
            if (this.auth) {
                yield this.headers.setAuthorization(this.auth);
            }
            try {
                return yield _super.execute.call(this);
            }
            catch (error) {
                if (retry) {
                    // Received an InvalidCredentialsError
                    if (error instanceof InvalidCredentialsError) {
                        if (isRefreshTokenRequestInProgress()) {
                            return REQUEST_QUEUE.add(() => {
                                const request = new KinveyHttpRequest(this);
                                return request.execute(false).catch(() => Promise.reject(error));
                            });
                        }
                        // Mark refresh token request in progress
                        markRefreshTokenRequestInProgress();
                        // Get existing mic session
                        const micSession = yield getKinveyMICSession();
                        if (micSession) {
                            try {
                                const isSuccess = yield this._refreshMICSession(micSession);
                                if (isSuccess) {
                                    // Redo the original request
                                    const request = new KinveyHttpRequest(this);
                                    const response = yield request.execute(false);
                                    // Mark the refresh token as complete
                                    markRefreshTokenRequestComplete();
                                    // Return the response
                                    return response;
                                }
                            }
                            catch (error) {
                                logger.error(error.message);
                            }
                        }
                        else {
                            yield cleanUp();
                        }
                        // Mark the refresh token as complete
                        markRefreshTokenRequestComplete();
                    }
                    // Throw the error
                    throw error;
                }
            }
        });
    }
}
//# sourceMappingURL=request.js.map