var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    function adopt(value) { return value instanceof P ? value : new P(function (resolve) { resolve(value); }); }
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : adopt(result.value).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
import { isArray, sum } from 'lodash-es';
import PQueue from 'p-queue';
import { ConfigKey, getConfig } from './config';
import { Query } from './query';
import { Aggregation } from './aggregation';
import { KinveyError } from './errors/kinvey';
import { getKey as getSessionKey } from './http';
const queue = new PQueue({ concurrency: 1 });
function generateId(length = 24) {
    const chars = 'abcdef0123456789';
    let id = '';
    for (let i = 0, j = chars.length; i < length; i += 1) {
        const pos = Math.floor(Math.random() * j);
        id += chars.substring(pos, pos + 1);
    }
    return id;
}
export class Storage {
    constructor(dbName, collectionName) {
        this.dbName = dbName;
        this.collectionName = collectionName;
    }
    get storageAdapter() {
        return getConfig(ConfigKey.StorageAdapter);
    }
    find(query) {
        return queue.add(() => __awaiter(this, void 0, void 0, function* () {
            if (query && !(query instanceof Query)) {
                throw new KinveyError('Invalid query. It must be an instance of the Query class.');
            }
            const docs = yield this.storageAdapter.find(this.dbName, this.collectionName);
            if (docs.length > 0 && query) {
                return query.process(docs);
            }
            return docs;
        }));
    }
    group(aggregation) {
        return __awaiter(this, void 0, void 0, function* () {
            if (!(aggregation instanceof Aggregation)) {
                throw new KinveyError('Invalid aggregation. It must be an instance of the Aggregation class.');
            }
            const docs = yield this.find();
            return aggregation.process(docs);
        });
    }
    count(query) {
        return __awaiter(this, void 0, void 0, function* () {
            if (query) {
                const docs = yield this.find(query);
                return docs.length;
            }
            return queue.add(() => this.storageAdapter.count(this.dbName, this.collectionName));
        });
    }
    findById(id) {
        return queue.add(() => this.storageAdapter.findById(this.dbName, this.collectionName, id));
    }
    save(docs) {
        return queue.add(() => __awaiter(this, void 0, void 0, function* () {
            let docsToSave = docs;
            let singular = false;
            if (!docs) {
                return null;
            }
            if (!isArray(docs)) {
                singular = true;
                docsToSave = [docs];
            }
            // Clone the docs
            docsToSave = docsToSave.slice(0, docsToSave.length);
            if (docsToSave.length > 0) {
                docsToSave = docsToSave.map((doc) => {
                    if (!doc._id) {
                        return Object.assign({}, {
                            _id: generateId(),
                            _kmd: Object.assign({}, doc._kmd, { local: true })
                        }, doc);
                    }
                    return doc;
                });
                yield this.storageAdapter.save(this.dbName, this.collectionName, docsToSave);
                return singular ? docsToSave.shift() : docsToSave;
            }
            return docs;
        }));
    }
    remove(query) {
        return __awaiter(this, void 0, void 0, function* () {
            const docs = yield this.find(query);
            if (query) {
                const results = yield Promise.all(docs.map(doc => this.removeById(doc._id)));
                return results.reduce((totalCount, count) => totalCount + count, 0);
            }
            yield this.clear();
            return docs.length;
        });
    }
    removeById(id) {
        return queue.add(() => this.storageAdapter.removeById(this.dbName, this.collectionName, id));
    }
    removeManyById(ids) {
        if (this.storageAdapter.removeManyById) {
            return queue.add(() => this.storageAdapter.removeManyById(this.dbName, this.collectionName, ids));
        }
        return Promise.all(ids.map(id => this.removeById(id)))
            .then(results => sum(results));
    }
    clear() {
        return queue.add(() => this.storageAdapter.clear(this.dbName, this.collectionName));
    }
    static clear(dbName) {
        const storageAdapter = getConfig(ConfigKey.StorageAdapter);
        const exclude = [getSessionKey()];
        return queue.add(() => storageAdapter.clearDatabase(dbName, exclude));
    }
}
//# sourceMappingURL=storage.js.map