var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    function adopt(value) { return value instanceof P ? value : new P(function (resolve) { resolve(value); }); }
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : adopt(result.value).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
import { isObjectLike, isBoolean } from 'lodash-es';
import { KinveyError } from '../errors/kinvey';
import { setSession, setMFASession, removeMFASession, formatKinveyBaasUrl, HttpRequestMethod, KinveyHttpRequest, KinveyBaasNamespace, KinveyHttpAuth, setDeviceToken, getDeviceToken, hasDeviceToken, removeDeviceToken, } from '../http';
import { User } from './user';
import { validateNoActiveUser, executeLoginRequest, validateCredentials } from './login';
const errMsgNoAuthenticators = 'User has no MFA authenticators yet. Please enable MFA.';
function executeChallengeRequest(authenticatorId) {
    return __awaiter(this, void 0, void 0, function* () {
        const request = new KinveyHttpRequest({
            method: HttpRequestMethod.POST,
            auth: KinveyHttpAuth.MFASessionToken,
            url: formatKinveyBaasUrl(KinveyBaasNamespace.User, '/mfa/challenge'),
            body: { authenticatorId },
        });
        const response = yield request.execute();
        return response.data;
    });
}
function executeCompleteRequest(code, trustDevice = false) {
    return __awaiter(this, void 0, void 0, function* () {
        const request = new KinveyHttpRequest({
            method: HttpRequestMethod.POST,
            auth: KinveyHttpAuth.MFASessionToken,
            url: formatKinveyBaasUrl(KinveyBaasNamespace.User, '/mfa/complete'),
            body: {
                code,
                createDeviceToken: trustDevice,
            },
        });
        const response = yield request.execute();
        response.data.user._kmd.authtoken = response.data.authToken;
        return response.data;
    });
}
function completeMFALoginRetryable(mfaComplete, context, maxRetriesCount) {
    return __awaiter(this, void 0, void 0, function* () {
        if (context.retries >= maxRetriesCount) {
            throw new KinveyError('Max retries count exceeded.');
        }
        const errMsgNoCode = 'MFA code is missing.';
        const errMsgTrustDevice = 'trustDevice should be boolean.';
        const mfaCompleteResult = yield mfaComplete(context.authenticator.id, context);
        if (!isObjectLike(mfaCompleteResult) || mfaCompleteResult.code == null) {
            throw new KinveyError(errMsgNoCode);
        }
        if (mfaCompleteResult.trustDevice != null && !isBoolean(mfaCompleteResult.trustDevice)) {
            throw new KinveyError(errMsgTrustDevice);
        }
        try {
            const trustDevice = mfaCompleteResult.trustDevice || false;
            const mfaData = yield executeCompleteRequest(mfaCompleteResult.code, trustDevice);
            return mfaData;
        }
        catch (err) {
            context.retries += 1; // eslint-disable-line no-param-reassign
            context.error = err; // eslint-disable-line no-param-reassign
            return completeMFALoginRetryable(mfaComplete, context, maxRetriesCount);
        }
    });
}
function _loginWithMFA(username, password, selectAuthenticator, mfaComplete, options = {}) {
    return __awaiter(this, void 0, void 0, function* () {
        yield validateNoActiveUser();
        const credentials = validateCredentials(username, password);
        if (!selectAuthenticator) {
            throw new KinveyError('Function to select authenticator is missing.');
        }
        if (!mfaComplete) {
            throw new KinveyError('Function to complete MFA is missing.');
        }
        const userHasDeviceToken = yield hasDeviceToken(credentials.username);
        if (userHasDeviceToken) {
            credentials.deviceToken = yield getDeviceToken(credentials.username);
        }
        const loginResult = yield executeLoginRequest(credentials, options.timeout);
        if (!loginResult.mfaRequired) {
            yield setSession(loginResult.user);
            return new User(loginResult.user);
        }
        if (userHasDeviceToken) {
            // MFA is still required which means that the device token has expired
            yield removeDeviceToken(credentials.username);
        }
        yield setMFASession({ mfaSessionToken: loginResult.mfaSessionToken, userId: loginResult.userId });
        if (loginResult.authenticators.length === 0) {
            throw new KinveyError(errMsgNoAuthenticators);
        }
        const context = {
            mfaSessionToken: loginResult.mfaSessionToken,
            retries: 0,
            authenticator: null,
        };
        const selectedAuthenticatorId = yield selectAuthenticator(loginResult.authenticators, context);
        if (selectedAuthenticatorId == null) {
            throw new KinveyError('MFA authenticator ID is missing.');
        }
        context.authenticator = loginResult.authenticators.find((a) => a.id === selectedAuthenticatorId);
        yield executeChallengeRequest(selectedAuthenticatorId);
        const mfaResult = yield completeMFALoginRetryable(mfaComplete, context, 10);
        if (mfaResult.deviceToken) {
            yield setDeviceToken(mfaResult.user.username, mfaResult.deviceToken);
        }
        yield removeMFASession();
        yield setSession(mfaResult.user);
        return new User(mfaResult.user);
    });
}
export function loginWithMFA(username, password, selectAuthenticator, mfaComplete, options = {}) {
    return __awaiter(this, void 0, void 0, function* () {
        try {
            return _loginWithMFA(username, password, selectAuthenticator, mfaComplete, options);
        }
        catch (err) {
            if (err.message !== errMsgNoAuthenticators) {
                yield removeMFASession();
            }
            throw err;
        }
    });
}
//# sourceMappingURL=loginWithMFA.js.map