var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    function adopt(value) { return value instanceof P ? value : new P(function (resolve) { resolve(value); }); }
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : adopt(result.value).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
import { isArray } from 'lodash-es';
import { Acl } from '../acl';
import { Kmd } from '../kmd';
import { getDeviceId } from '../device';
import { getSession, setSession, removeSession, removeMFASession, removeDeviceToken, formatKinveyBaasUrl, HttpRequestMethod, KinveyHttpRequest, KinveyBaasNamespace, KinveyHttpAuth, } from '../http';
import { KinveyError } from '../errors/kinvey';
import { DataStoreCache, QueryCache, SyncCache } from '../datastore/cache';
import { subscribe, unsubscribe, isSubscribed } from '../live';
import { logger } from '../log';
import { mergeSocialIdentity } from './utils';
import { signup } from './signup';
import { createMFAAuthenticator, } from './createMFAAuthenticator';
export class User {
    constructor(data = {}) {
        this.data = data;
    }
    get _id() {
        if (this.data) {
            return this.data._id;
        }
        return undefined;
    }
    get _acl() {
        if (this.data) {
            return new Acl(this.data);
        }
        return undefined;
    }
    get _kmd() {
        if (this.data) {
            return new Kmd(this.data);
        }
        return undefined;
    }
    get metadata() {
        return this._kmd;
    }
    get authtoken() {
        const kmd = this._kmd;
        if (kmd) {
            return kmd.authtoken;
        }
        return undefined;
    }
    get _socialIdentity() {
        return this.data._socialIdentity;
    }
    get username() {
        if (this.data) {
            return this.data.username;
        }
        return undefined;
    }
    get email() {
        if (this.data) {
            return this.data.email;
        }
        return undefined;
    }
    isActive() {
        return __awaiter(this, void 0, void 0, function* () {
            const activeUser = yield getSession();
            if (activeUser && activeUser._id === this._id) {
                return true;
            }
            return false;
        });
    }
    isEmailVerified() {
        const metadata = this.metadata;
        if (metadata) {
            return metadata.isEmailConfirmed();
        }
        return false;
    }
    signup(options = {}) {
        return __awaiter(this, void 0, void 0, function* () {
            return signup(this.data, options);
        });
    }
    me(options = {}) {
        return __awaiter(this, void 0, void 0, function* () {
            const request = new KinveyHttpRequest({
                method: HttpRequestMethod.GET,
                auth: KinveyHttpAuth.Session,
                url: formatKinveyBaasUrl(KinveyBaasNamespace.User, '/_me'),
                timeout: options.timeout,
            });
            const response = yield request.execute();
            const { data } = response;
            // Remove sensitive data
            delete data.password;
            // Merge _socialIdentity
            if (data._socialIdentity) {
                this.data = yield getSession();
                if (data._socialIdentity.kinveyAuth) {
                    delete data._socialIdentity.kinveyAuth.refresh_token;
                }
                data._socialIdentity = mergeSocialIdentity(this._socialIdentity, data._socialIdentity);
            }
            // Update the active session
            if (yield this.isActive()) {
                data._kmd.authtoken = this.authtoken;
                yield setSession(data);
            }
            this.data = data;
            return this;
        });
    }
    update(data, options = {}) {
        return __awaiter(this, void 0, void 0, function* () {
            const body = Object.assign({}, this.data, data);
            if (!data) {
                throw new KinveyError('No user was provided to be updated.');
            }
            if (isArray(data)) {
                throw new KinveyError('Only one user can be updated at one time.');
            }
            if (!body._id) {
                throw new KinveyError('User must have an _id.');
            }
            const request = new KinveyHttpRequest({
                method: HttpRequestMethod.PUT,
                auth: KinveyHttpAuth.SessionOrMaster,
                url: formatKinveyBaasUrl(KinveyBaasNamespace.User, `/${this._id}`),
                body,
                timeout: options.timeout
            });
            const response = yield request.execute();
            const updatedData = response.data;
            // Remove sensitive data
            delete updatedData.password;
            // Merge _socialIdentity
            if (updatedData._socialIdentity) {
                updatedData._socialIdentity = mergeSocialIdentity(this._socialIdentity, updatedData._socialIdentity);
            }
            // Update the active session
            if (yield this.isActive()) {
                yield setSession(updatedData);
            }
            this.data = updatedData;
            return this;
        });
    }
    registerForLiveService(options = {}) {
        return __awaiter(this, void 0, void 0, function* () {
            if (!isSubscribed()) {
                // Register the user
                const deviceId = yield getDeviceId();
                const request = new KinveyHttpRequest({
                    method: HttpRequestMethod.POST,
                    auth: KinveyHttpAuth.Session,
                    url: formatKinveyBaasUrl(KinveyBaasNamespace.User, `/${this._id}/register-realtime`),
                    body: { deviceId },
                    timeout: options.timeout
                });
                const response = yield request.execute();
                const config = Object.assign({}, { authKey: this.authtoken, uuid: deviceId }, response.data);
                // Subscribe to PubNub
                subscribe(config);
            }
            return true;
        });
    }
    unregisterFromLiveService(options = {}) {
        return __awaiter(this, void 0, void 0, function* () {
            if (isSubscribed()) {
                // Unsubscribe from PubNub
                unsubscribe();
                // Unregister the user
                const deviceId = yield getDeviceId();
                const request = new KinveyHttpRequest({
                    method: HttpRequestMethod.POST,
                    auth: KinveyHttpAuth.Session,
                    url: formatKinveyBaasUrl(KinveyBaasNamespace.User, `/${this._id}/unregister-realtime`),
                    body: { deviceId },
                    timeout: options.timeout
                });
                yield request.execute();
            }
            return true;
        });
    }
    createAuthenticator(newAuthenticator, verify) {
        return __awaiter(this, void 0, void 0, function* () {
            return createMFAAuthenticator(this._id, newAuthenticator, verify);
        });
    }
    listAuthenticators() {
        return __awaiter(this, void 0, void 0, function* () {
            const request = new KinveyHttpRequest({
                method: HttpRequestMethod.GET,
                auth: KinveyHttpAuth.SessionOrMaster,
                url: formatKinveyBaasUrl(KinveyBaasNamespace.User, `/${this._id}/authenticators`),
            });
            const { data } = yield request.execute();
            return data.map((a) => {
                return { id: a.id, name: a.name, type: a.type };
            });
        });
    }
    removeAuthenticator(id) {
        return __awaiter(this, void 0, void 0, function* () {
            const request = new KinveyHttpRequest({
                method: HttpRequestMethod.DELETE,
                auth: KinveyHttpAuth.SessionOrMaster,
                url: formatKinveyBaasUrl(KinveyBaasNamespace.User, `/${this._id}/authenticators/${id}/`),
            });
            yield request.execute();
            return null;
        });
    }
    listRecoveryCodes() {
        return __awaiter(this, void 0, void 0, function* () {
            const request = new KinveyHttpRequest({
                method: HttpRequestMethod.GET,
                auth: KinveyHttpAuth.SessionOrMaster,
                url: formatKinveyBaasUrl(KinveyBaasNamespace.User, `/${this._id}/recovery-codes`),
            });
            const { data } = yield request.execute();
            return data.recoveryCodes;
        });
    }
    regenerateRecoveryCodes() {
        return __awaiter(this, void 0, void 0, function* () {
            const request = new KinveyHttpRequest({
                method: HttpRequestMethod.POST,
                auth: KinveyHttpAuth.SessionOrMaster,
                url: formatKinveyBaasUrl(KinveyBaasNamespace.User, `/${this._id}/recovery-codes`),
            });
            const { data } = yield request.execute();
            return data.recoveryCodes;
        });
    }
    isMFAEnabled() {
        return __awaiter(this, void 0, void 0, function* () {
            return (yield this.listAuthenticators()).length > 0;
        });
    }
    disableMFA() {
        return __awaiter(this, void 0, void 0, function* () {
            const authenticators = yield this.listAuthenticators();
            yield Promise.all(authenticators.map((a) => this.removeAuthenticator(a.id)));
            return true;
        });
    }
    _cleanup(kinveyRequest, operationName, cleanEntireSessionStore = false) {
        return __awaiter(this, void 0, void 0, function* () {
            if (!(yield this.isActive())) {
                return this;
            }
            this.unregisterFromLiveService();
            // TODO: unregister push
            try {
                yield kinveyRequest.execute();
            }
            catch (error) {
                logger.error(`${operationName} failed.`);
                logger.error(error.message);
            }
            yield removeSession();
            if (cleanEntireSessionStore) {
                yield removeMFASession();
                yield removeDeviceToken(this.data.username);
            }
            yield QueryCache.clear();
            yield SyncCache.clear();
            yield DataStoreCache.clear();
            return this;
        });
    }
    logout(options = {}) {
        return __awaiter(this, void 0, void 0, function* () {
            const request = new KinveyHttpRequest({
                method: HttpRequestMethod.POST,
                auth: KinveyHttpAuth.Session,
                url: formatKinveyBaasUrl(KinveyBaasNamespace.User, '/_logout'),
                timeout: options.timeout,
            });
            return this._cleanup(request, 'Logout request');
        });
    }
    invalidateTokens() {
        return __awaiter(this, void 0, void 0, function* () {
            const request = new KinveyHttpRequest({
                method: HttpRequestMethod.DELETE,
                auth: KinveyHttpAuth.Session,
                url: formatKinveyBaasUrl(KinveyBaasNamespace.User, `/${this._id}/tokens`),
            });
            return this._cleanup(request, 'Tokens invalidation', true);
        });
    }
}
//# sourceMappingURL=user.js.map